"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const luxon_1 = require("luxon");
const express_1 = __importDefault(require("express"));
const express_async_handler_1 = __importDefault(require("express-async-handler"));
const client_1 = require("@taskbotjs/client");
const APIError_1 = require("./APIError");
const Middleware = __importStar(require("./Middleware"));
class WebAPIApp {
    constructor(baseLogger, clientPool, logRequests = true) {
        this.clientPool = clientPool;
        this.logger = baseLogger.child({ component: "WebAPIApp" });
        this.expressApp = this.buildExpress(logRequests);
    }
    buildExpress(logRequests = true) {
        const logger = this.logger.child({ component: "Express" });
        const app = express_1.default();
        if (logRequests) {
            app.use(Middleware.buildLogRequestsMiddleware(logger));
        }
        app.use(Middleware.paginationMiddleware);
        app.get("/ping", (req, res) => {
            const resp = { pong: true };
            res.json(resp);
        });
        app.get("/workers", express_async_handler_1.default(async (req, res) => {
            const resp = await this.clientPool.use((taskbot) => taskbot.getWorkerInfo());
            res.json(resp);
        }));
        app.post("/workers/cleanup", express_async_handler_1.default(async (req, res) => {
            await this.clientPool.use((taskbot) => taskbot.cleanUpDeadWorkers());
            res.json({ ok: true });
        }));
        app.get("/queues", express_async_handler_1.default(async (req, res) => {
            const resp = await this.clientPool.use((taskbot) => taskbot.getQueueInfo());
            res.json(resp);
        }));
        app.get("/queues/:queueName", express_async_handler_1.default(async (req, res) => {
            const queueName = req.params.queueName;
            const limit = req.query.limit;
            const offset = req.query.offset;
            const resp = await this.clientPool.use((taskbot) => taskbot.withQueue(queueName, (queue) => queue.peek(limit, offset)));
            res.json(resp);
        }));
        app.get("/queues/:queueName/:jobId", express_async_handler_1.default(async (req, res) => {
            const queueName = req.params.queueName;
            const jobId = req.params.jobId;
            const resp = await this.clientPool.use(async (taskbot) => {
                const queue = taskbot.queue(queueName);
                return queue.find((jd) => jd.id === jobId);
            });
            if (!resp) {
                throw new APIError_1.NotFoundError(`Job '${jobId}' not found in queue. It may have already been dequeued.`);
            }
            res.json(resp);
        }));
        app.delete("/queues/:queueName/:jobId", express_async_handler_1.default(async (req, res) => {
            const queueName = req.params.queueName;
            const jobId = req.params.jobId;
            await this.clientPool.use((taskbot) => taskbot.withQueue(queueName, (queue) => queue.remove(jobId)));
            res.json({ ok: true });
        }));
        app.post("/queues/:queueName/:jobId/launch", express_async_handler_1.default(async (req, res) => {
            const queueName = req.params.queueName;
            const jobId = req.params.jobId;
            await this.clientPool.use((taskbot) => taskbot.withQueue(queueName, async (queue) => {
                if (!await queue.launch(jobId)) {
                    throw new APIError_1.NotFoundError(`Job '${jobId}' not pushed to head. It may have already been processed.`);
                }
            }));
            res.json({ ok: true });
        }));
        app.get("/metrics/basic", express_async_handler_1.default(async (req, res) => {
            res.json(await this.clientPool.use(async (taskbot) => taskbot.getBasicMetrics()));
        }));
        app.get("/metrics/dated", express_async_handler_1.default(async (req, res) => {
            const now = luxon_1.DateTime.utc();
            let start = now;
            let end = now;
            if (req.query.start) {
                start = luxon_1.DateTime.fromFormat(req.query.start, client_1.LUXON_YMD, { zone: "UTC" });
            }
            if (req.query.end) {
                end = luxon_1.DateTime.fromFormat(req.query.end, client_1.LUXON_YMD, { zone: "UTC" });
            }
            const resp = await this.clientPool.use(async (taskbot) => taskbot.getDatedMetrics(start, end));
            res.json(resp);
        }));
        app.get("/metrics/storage", express_async_handler_1.default(async (req, res) => {
            const resp = await this.clientPool.use(async (taskbot) => taskbot.getStorageMetrics());
            res.json(resp);
        }));
        app.get("/scheduled", express_async_handler_1.default(async (req, res) => {
            const resp = await this.clientPool.use((taskbot) => taskbot.scheduleSet.peek(req.query.limit, req.query.offset));
            res.json(resp);
        }));
        app.get("/scheduled/:jobId", express_async_handler_1.default(async (req, res) => {
            const jobId = req.params.jobId;
            const resp = await this.clientPool.use(async (taskbot) => {
                if (await taskbot.scheduleSet.contains(jobId)) {
                    return taskbot.readJob(jobId);
                }
                return null;
            });
            if (!resp) {
                throw new APIError_1.NotFoundError(`Job '${jobId}' not found.`);
            }
            else {
                res.json(resp);
            }
        }));
        app.post("/scheduled/:jobId/launch", express_async_handler_1.default(async (req, res) => {
            const jobId = req.params.jobId;
            const ret = await this.clientPool.use((taskbot) => taskbot.scheduleSet.launch(jobId));
            if (!ret) {
                throw new APIError_1.NotFoundError(`Job '${jobId}' not found. Possibly already out of the set.`);
            }
            res.json({ ok: true });
        }));
        app.delete("/scheduled/:jobId", express_async_handler_1.default(async (req, res) => {
            const jobId = req.params.jobId;
            await this.clientPool.use(async (taskbot) => {
                await taskbot.scheduleSet.remove(jobId);
                await taskbot.unsafeDeleteJob(jobId);
            });
            res.json({ ok: true });
        }));
        app.get("/retry", express_async_handler_1.default(async (req, res) => {
            const resp = await this.clientPool.use((taskbot) => taskbot.retrySet.peek(req.query.limit, req.query.offset));
            res.json(resp);
        }));
        app.get("/retry/:jobId", express_async_handler_1.default(async (req, res) => {
            const jobId = req.params.jobId;
            const resp = await this.clientPool.use(async (taskbot) => {
                if (await taskbot.retrySet.contains(jobId)) {
                    return taskbot.readJob(jobId);
                }
                return null;
            });
            if (!resp) {
                throw new APIError_1.NotFoundError(`Job '${jobId}' not found.`);
            }
            else {
                res.json(resp);
            }
        }));
        app.post("/retry/:jobId/launch", express_async_handler_1.default(async (req, res) => {
            const jobId = req.params.jobId;
            const ret = await this.clientPool.use((taskbot) => taskbot.retrySet.retry(jobId));
            if (!ret) {
                throw new APIError_1.NotFoundError(`Job '${jobId}' not found. Possibly already out of the set.`);
            }
            res.json({ ok: true });
        }));
        app.delete("/retry/:jobId", express_async_handler_1.default(async (req, res) => {
            const jobId = req.params.jobId;
            await this.clientPool.use(async (taskbot) => {
                await taskbot.retrySet.remove(jobId);
                await taskbot.unsafeDeleteJob(jobId);
            });
            res.json({ ok: true });
        }));
        app.get("/dead", express_async_handler_1.default(async (req, res) => {
            const resp = await this.clientPool.use((taskbot) => taskbot.deadSet.peek(req.query.limit, req.query.offset));
            res.json(resp);
        }));
        app.post("/dead/clean", express_async_handler_1.default(async (req, res) => {
            const count = await this.clientPool.use((taskbot) => taskbot.deadSet.cleanAll());
            res.json({ ok: true, count });
        }));
        app.get("/dead/:jobId", express_async_handler_1.default(async (req, res) => {
            const jobId = req.params.jobId;
            const resp = await this.clientPool.use(async (taskbot) => {
                if (await taskbot.deadSet.contains(jobId)) {
                    return taskbot.readJob(jobId);
                }
                return null;
            });
            if (!resp) {
                throw new APIError_1.NotFoundError(`Job '${jobId}' not found.`);
            }
            else {
                res.json(resp);
            }
        }));
        app.post("/dead/:jobId/launch", express_async_handler_1.default(async (req, res) => {
            const jobId = req.params.jobId;
            const ret = await this.clientPool.use((taskbot) => taskbot.deadSet.resurrect(jobId));
            if (!ret) {
                throw new APIError_1.NotFoundError(`Job '${jobId}' not found. Possibly already out of the set.`);
            }
            res.json({ ok: true });
        }));
        app.delete("/dead/:jobId", express_async_handler_1.default(async (req, res) => {
            const jobId = req.params.jobId;
            await this.clientPool.use(async (taskbot) => {
                await taskbot.deadSet.remove(jobId);
                await taskbot.unsafeDeleteJob(jobId);
            });
            res.json({ ok: true });
        }));
        if (logRequests) {
            app.use(Middleware.buildErrorLoggingMiddleware(logger));
        }
        return app;
    }
}
exports.WebAPIApp = WebAPIApp;
